#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "project3.h"
#include "sharedcode.h"

extern int TraceLevel;
extern float clocktime;
#define DESIRED_TRACE -1

/* * * Generic functions used by all nodes * * */

/* initialize_DistanceTable
        Initializa the given distance table */
void initialize_DistanceTable(int nodeNum, struct distance_table *ptrDtx,
        struct NeighborCosts **ptrNC){
    int i,j;
    
    // Pre initialize table by setting all values to INFINITY:
    for ( i = 0; i < MAX_NODES; i++ ) {
        for ( j = 0; j < MAX_NODES; j++ ) {
            ptrDtx->costs[i][j] = INFINITY;
        }
    }

    *ptrNC = getNeighborCosts(nodeNum);

    // Edit the distance table with the cost to your neigbhor:
    for (j = 0; j < MAX_NODES; j++){
        ptrDtx->costs[j][j] = (*ptrNC)->NodeCosts[j];
    }
}

/* calculate_MinCost
        Calculate the minimum path cost using the given distance table */
void calculate_MinCost(int nodeNum, struct minimum_cost_strct *ptrMCS, 
        struct distance_table *ptrDtx){
   int i, j;

    // Calculate minimum cost
    for (j = 0; j < MAX_NODES; j++){
        for (i = 0; i < MAX_NODES; i++){
            if (ptrDtx->costs[j][i] < ptrMCS->mincost[j]) {
                ptrMCS->mincost[j] = ptrDtx->costs[j][i];
            }
        }
    }
}

/* makeNsend_RoutePackets
       Create RoutPacket for each neighbor and send to each neighbor*/
void makeNsend_RoutePackets(int nodeNum, struct NeighborCosts *ptrNC, struct minimum_cost_strct *ptrMCS,
        struct distance_table *ptrDtx){
    int i,j;

    // Make and send a packet for each neighbor:
    for (i = 0; i < MAX_NODES; i++){
        if (ptrNC->NodeCosts[i] != 0 && ptrNC->NodeCosts[i] != INFINITY){
            struct RoutePacket *pkt2send = (struct RoutePacket *) malloc(sizeof(struct RoutePacket));

            pkt2send->sourceid = nodeNum; // Assign source.
            pkt2send->destid = i; // Assign destination.

            // Fill the packet with minimum cost:
            for (j = 0; j < MAX_NODES; j++){
                pkt2send->mincost[j] = ptrMCS->mincost[j];
            }

            // Poision Reverse advertised minimum costs:
            for (j = 0; j < MAX_NODES; j++){
                // If my lowest cost to j is equal to my cost to j via you...
                if (ptrMCS->mincost[j] == ptrDtx->costs[j][pkt2send->destid] && ptrMCS->mincost[j] != 0){
                    pkt2send->mincost[j] = INFINITY; // Advertise that their's no connection.
                }
            }

            // Send packet to neighbor and print information:
            toLayer2(*pkt2send);
            printf("At time t=%f, node %d, sends packet to node %d with: %d %d %d %d\n",
                clocktime, pkt2send->sourceid, pkt2send->destid, pkt2send->mincost[0], 
                pkt2send->mincost[1], pkt2send->mincost[2], pkt2send->mincost[3]);
        }
    }
}

/* update_function
       update the distance table and minimum cost
       return 0 if the minimum cost was updated; -1 if mincost wasn't updated */
int update_function(int nodeNum, struct minimum_cost_strct *ptrMCS, 
        struct distance_table *ptrDtx, struct RoutePacket *rcvdpkt){
    int j;
    int srcNd = rcvdpkt->sourceid;
    int minUpdated = -1; // -1 == false : 0 == true

    printf("Recieved update from %d its min cost is: %d %d %d %d\n", rcvdpkt->sourceid,
            rcvdpkt->mincost[0], rcvdpkt->mincost[1], rcvdpkt->mincost[2], rcvdpkt->mincost[3]);

    printf("My Initial min cost is: %d %d %d %d\n",
            ptrMCS->mincost[0], ptrMCS->mincost[1], ptrMCS->mincost[2], ptrMCS->mincost[3]);

    for (j = 0; j < MAX_NODES; j++) {
        if (j == nodeNum || j == srcNd){
            /* ignore checking neighbors min cost to myself
                and ignore checking neighbors cost to itself */
        } else if (rcvdpkt->mincost[j] == (ptrDtx->costs[srcNd][srcNd] + ptrMCS->mincost[j])) {
            /* ignore If updated path includes going back into your own node */
        } else if ((ptrDtx->costs[srcNd][srcNd] + rcvdpkt->mincost[j]) < ptrMCS->mincost[j]) {
            minUpdated = 0;
            // update my mincost:
            ptrMCS->mincost[j] = ptrDtx->costs[srcNd][srcNd] + rcvdpkt->mincost[j];
            // update my distance table:
            ptrDtx->costs[j][srcNd] = ptrDtx->costs[srcNd][srcNd] + rcvdpkt->mincost[j];

        } else if (rcvdpkt->mincost[j] != INFINITY){
            /* if there is no new optimal route, update the distance table with poor route */
            ptrDtx->costs[j][srcNd] = ptrDtx->costs[srcNd][srcNd] + rcvdpkt->mincost[j];
        }
    }

    return minUpdated;
}

/* tracePrint 
     print msg if Tracelevel > 2. */
void tracePrint(char *msg, int isTimeDesired){
    if (isTimeDesired == 1){
        if (TraceLevel > DESIRED_TRACE){
        /* if time is desired print message with timestamp */
            printf("%s%f.\n", msg, clocktime);
        }
    } else {
        if (TraceLevel > DESIRED_TRACE){
        /* else just print the trace message */
            printf("%s\n", msg);
        }
    }
}

/* error
     Print error with the given message */
void error(char *msg){ // Print the given error message.
    fprintf(stderr, "%s\n", msg);
    exit(-1);
}
















